<?php defined('SYSPATH') OR die('No direct access allowed.');

/**
 * Handle comments for hosts and services
 */
class Comment_Model extends Model {
	/***************************** COMMENT TYPES *******************************/
	const HOST_COMMENT = 1; /**< Comment applies to host */
	const SERVICE_COMMENT = 2; /**< Comment applies to service */

	/****************************** ENTRY TYPES ********************************/
	const USER_COMMENT = 1; /**< Comment is entered by user */
	const DOWNTIME_COMMENT = 2; /**< Comment is generated by a downtime */
	const FLAPPING_COMMENT = 3; /**< Comment is generated because object is flapping */
	const ACKNOWLEDGEMENT_COMMENT = 4; /**< Comment is generated from an acknowledgement */

	const TABLE_NAME = 'comments'; /**< The name of the comment livestatus table */

	
	/**
	 * Fetch saved comments for host or service
	 *
	 * @param $host string Host name - must be set
	 * @param $service string Service description, or false to get comments for a host
	 * @param $num_per_page int Number of rows to retrieve
	 * @param $offset int Number of rows to skip before retrieving first row
	 * @param $count bool Completely ignore the two previous options - return the total number of comments instead of the comments themselves
	 * @returns If $count is true, then the number of rows as an int, otherwise the database result of comments
	 */
	public static function fetch_comments_by_object($host=false, $service=false, $num_per_page=false, $offset=false, $count=false)
	{
		$num_per_page = (int)$num_per_page;
		$offset = (int)$offset;
		if (empty($host)) {
			return false;
		}
		$ls = Livestatus::instance();
		$lsb = $ls->getBackend();
		
		if( $service == false ) {
			$filter = array('service_description' => '', 'host_name' => $host);
		} else {
			$filter = array('host_name' => $host, 'service_description' => $service);
		}
		if( $count ) {
			$result = $lsb->getStats(static::TABLE_NAME, array(
					'count' => array('author' => array( '!=' => '' ))
					), array(
					'filter' => $filter
					) );
			return $result[0]['count'];
		}
		if( static::TABLE_NAME == 'comments' ) {
			$result = $ls->getComments( array(
					'filter' => $filter,
					) );
		} else {
			$result = $ls->getDowntimes( array(
					'filter' => $filter,
					) );
		}
		
		return $result;
	}

	/**
	 * Fetch all host or all service comments the current user may see
	 *
	 * @param $for_services bool If true, fetch service comments, else fetch host comments
	 * @param $num_per_page int Number of rows to retrieve
	 * @param $offset int Number of rows to skip before retrieving first row
	 * @param $count bool Completely ignore the two previous options - return the total number of comments instead of the comments themselves
	 * @returns If $count is true, then the number of rows as an int, otherwise the database result of comments
	 */
	public static function fetch_comments_by_user($for_services=false, $num_per_page=false, $offset=false, $count=false)
	{
		$num_per_page = (int)$num_per_page;
		$offset = (int)$offset;

		$ls = Livestatus::instance();
		$lsb = $ls->getBackend();

		if ($for_services) {
			$filter = array('service_description' => array('!=' => ''));
		} else {
			$filter = array('service_description' => '');
		}
		if ($count) {
			$result = $lsb->getStats(static::TABLE_NAME, array(
					'count' => array('author' => array( '!=' => '' ))
					), array(
					'filter' => $filter
					) );
			return $result[0]['count'];
		}

		$options = array(
			'filter' => $filter
		);
		if($num_per_page) {
			$options['limit'] = $num_per_page;
		}
		if($offset) {
			$options['offset'] = $offset;
		}
		if (static::TABLE_NAME == 'comments') {
			$result = $ls->getComments($options);
		} else {
			$result = $ls->getDowntimes($options);
		}

		return $result;
	}

	/**
	 * Fetch all comments of a specified type for a specified host or service
	 * @param $entry_type The comment type as an integer
	 * @param $host_name Host name
	 * @param $service_description Service description, or empty to fetch for host
	 * @return DB result, or false on error or empty
	 */
	public static function fetch_all_comment_types($entry_type, $host_name, $service_description) {

		$db = Database::instance();
		switch ($entry_type) {
			case 1: // user comment
			case 3: // flapping
			case 4: // acknowledged
				$type = static::TABLE_NAME;
				break;
			case 2: // downtime
				$type = 'scheduled_downtime';
				break;
			default:
				die("What did you do!?");
				break;
		}
		$and = empty($service_description) ? '' : " AND service_description='".$service_description."'";
		$sql = "SELECT comment_data from ".$type." where host_name = '".$host_name."'".$and."";
		$result = $db->query($sql);

		return $result->count() ? $result->result() : false;
	}

	/**
	*	Wrapper method to fetch nr of comments for host or service
	*/
	public static function count_comments_by_object($host=false, $service=false)
	{
		return Comment_Model::fetch_comments_by_object($host, $service, false, false, true);
	}

	/**
	*	Wrapper method to fetch a count of all service- or host comments
	*/
	public static function count_comments_by_user($host=false, $service=false)
	{
		return static::fetch_comments_by_user($service, false, false, true);
	}

	/**
	*	Fetch comment counts for all objects that has comments
	*	Returned array will contain object name as key and count
	* 	as value for all objects with comments.
	*/
	public static function count_all_comments_by_object($service=false)
	{
		$ls = Livestatus::instance();
		if( $service ) {
			$comments = $ls->getBackend()->getStats( 'comments',
					array(
							'count' => array('service_state' => array( '!=' => 999 ))
							),
					array(
							'columns' => array('service_description')
							)
					);
			$name_column = 'service_description';
		} else {
			$comments = $ls->getBackend()->getStats( 'comments',
					array(
							'count' => array('host_name' => array( '!=' => '' ))
							),
					array(
							'columns' => array('host_name')
							)
					);
			$name_column = 'host_name';
		}
		
		$data = array();
		foreach( $comments as $row ) {
			$data[$row[$name_column]] = $row['count'];
		}
		if( empty( $data ) ) {
			return false;
		}
		
		return $data;
	}

	/**
	*	Search through several fields for a specific value
	*/
	public function search($value=false, $limit=false)
	{
		if (empty($value)) return false;
		$db = Database::instance();
		$auth = Nagios_auth_Model::instance();
		$contact_id = (int)$auth->id;
		$limit_str = sql::limit_parse($limit);
		$join_host = false;
		$where_host = false;
		$join_svc = false;
		$where_svc = false;
		if (!$auth->view_hosts_root) {
			$join_host = "INNER JOIN contact_access ON host.id = contact_access.host ";
			$where_host = "AND contact_access.contact = ".$contact_id." ";
			$join_svc = "INNER JOIN contact_access ON service.id = contact_access.service ";
			$where_svc = "AND contact_access.contact = ".$contact_id." ".
				"AND service.host_name = host.host_name ";
		}
		if (is_array($value) && !empty($value)) {
			$query = false;
			$sql = false;
			foreach ($value as $val) {
				$val = '%'.$val.'%';
				$query[] = "SELECT c.id FROM ".static::TABLE_NAME." c ".
				" INNER JOIN host on host.host_name = c.host_name ".$join_host.
				" WHERE (LCASE(comment_data) LIKE LCASE(".$db->escape($val).") OR ".
				"LCASE(c.host_name) LIKE LCASE(".$db->escape($val).") ) ".
				" AND c.service_description IS NULL ".$where_host.
				" UNION ".
				"SELECT c.id FROM ".static::TABLE_NAME." c ".
				"INNER JOIN host on host.host_name = c.host_name ".
				"INNER JOIN service ON service.service_description = c.service_description ".$join_svc.
				" WHERE (LCASE(comment_data) LIKE LCASE(".$db->escape($val).") OR ".
				"LCASE(c.host_name) LIKE LCASE(".$db->escape($val).") OR ".
				"LCASE(c.service_description) LIKE LCASE(".$db->escape($val).") ) ".
				" AND c.service_description IS NOT NULL ".
				"AND service.host_name = host.host_name ".$where_svc;
			}
			if (!empty($query)) {
				$sql = 'SELECT * FROM '.static::TABLE_NAME.' WHERE id IN ('.implode(' UNION ', $query).') ORDER BY host_name, service_description, entry_time '.$limit_str;
			}
		} else {
			$value = '%'.$value.'%';
			$sql = "(SELECT c.* FROM ".static::TABLE_NAME." c ".
				" INNER JOIN host on host.host_name = c.host_name ".$join_host.
				"WHERE (LCASE(comment_data) LIKE LCASE(".$db->escape($value).") OR ".
				"LCASE(c.host_name) LIKE LCASE(".$db->escape($value).") )".
				" AND c.service_description IS NULL ".$where_host.
				") UNION ALL (".
				"SELECT c.* FROM ".static::TABLE_NAME." c ".
				"INNER JOIN host on host.host_name = c.host_name ".
				"INNER JOIN service ON service.service_description = c.service_description ".$join_svc.
				"WHERE (LCASE(comment_data) LIKE LCASE(".$db->escape($value).") OR ".
				"LCASE(c.host_name) LIKE LCASE(".$db->escape($value).") OR ".
				"LCASE(c.service_description) LIKE LCASE(".$db->escape($value).") ) ".
				"AND c.service_description IS NOT NULL ".
				"AND service.host_name = host.host_name ".$where_svc." )".$limit_str;
		}
		$obj_info = $db->query($sql);
		return $obj_info;
	}

	/**
	*	Fetch comment info filtered on specific field and value
	*/
	public function get_where($field=false, $value=false, $limit=false)
	{
		if (empty($field) || empty($value)) {
			return false;
		}
		$db = Database::instance();
		$auth = Nagios_auth_Model::instance();
		$field = trim($field);
		$value = trim($value);
		$contact_id = (int)$auth->id;
		$limit_str = sql::limit_parse($limit);
		$join_host = false;
		$where_host = false;
		$join_svc = false;
		$where_svc = false;
		if (!$auth->view_hosts_root) {
			$join_host = "INNER JOIN contact_access ON host.id = contact_access.host ";
			$where_host = "AND contact_access.contact = ".$contact_id;
			$join_svc = "INNER JOIN contact_access ON service.id = contact_access.service ";
			$where_svc = "AND contact_access.contact = ".$contact_id." AND service.host_name = host.host_name ";;
		}

		$limit_str = sql::limit_parse($limit);
		$value = '%' . $value . '%';
		$sql = "(SELECT c.* FROM ".static::TABLE_NAME." c ".
			" INNER JOIN host on host.host_name = c.host_name ".$join_host.
			"WHERE LCASE(".$field.") LIKE LCASE(".$db->escape($value).")".
			" AND c.service_description IS NULL ".$where_host.
			") UNION ALL (".
			"SELECT c.* FROM ".static::TABLE_NAME." c ".
			"INNER JOIN host on host.host_name = c.host_name ".
			"INNER JOIN service ON service.service_description = c.service_description ".$join_svc.
			"WHERE LCASE(".$field.") LIKE LCASE(".$db->escape($value).") ".
			"AND c.service_description IS NOT NULL ".
			"AND service.host_name = host.host_name ".$where_svc." )".$limit_str;
		$obj_info = $db->query($sql);
		return count($obj_info) > 0 ? $obj_info : false;
	}

}
