<?php

/**
 * Generate test coverage output.
 * Use like this:
 *     xdebug_start_code_coverage(XDEBUG_CC_UNUSED | XDEBUG_CC_DEAD_CODE);
 *     actual_tests()
 *     exit(generate_coverage(xdebug_get_code_coverage(),
 *                            array('testsuite/path', 'vendor/library/path'));
 * Tests will be written to '/var/www/coverage/', with a index.html overview in
 * the main dir, and all other files in their respective sub-directories.
 * Set $output_prefix to FALSE to disable generating HTML files.
 *
 * Uses the environment variables WARNING and CRITICAL if set to return proper
 * nagios exit code. Uses the environment variable COVERAGE to decide whether
 * to generate HTML output, and if so, where.
 */
function generate_coverage($coverage, $exclude_dirs=array()) {
	$warning = getenv('WARNING');
	$critical = getenv('CRITICAL');
	$output_prefix = getenv('COVERAGE');
	$all_files = array();
	$total_lines = 0;
	$total_covered = 0;
	$completely_tested = array();
	foreach ($coverage as $file => $lines) {
		// we don't really care about assert or eval code - at least for now.
		if (strrpos($file, 'assert code') !== false)
			continue;
		if (strrpos($file, 'eval()\'d code') !== false)
			continue;
		// relative path plz
		$dir = getcwd();
		if (strncmp($file, $dir, strlen($dir)) === 0)
			$file = substr($file, strlen($dir)+1);

		foreach ($exclude_dirs as $dir) {
			if (strncmp($file, $dir, strlen($dir)) === 0)
				continue 2;
		}
		$in = fopen($file, 'r');
		if ($output_prefix !== false) {
			@mkdir($output_prefix.'/'.dirname($file), 0777, true);
			$out = fopen($output_prefix.'/'.$file.'.html', 'w');
			fwrite($out, '<html><head><style type="text/css">.visited{color: #0c0;} .unused{color: #600;} .dead{color: #f00;} li{font-family: monospace;}</style></head><body><ol>');
		}
		$visited=0;
		$unused=0;
		$dead=0;
		$i = 1;
		while ($line = fgets($in)) {
			if (!isset($lines[$i])) {
				$class = 'normal';
			} elseif ($lines[$i] === 1) {
				$visited++;
				$class = 'visited';
			} elseif ($lines[$i] === -1) {
				$unused++;
				$class = 'unused';
			} elseif ($lines[$i] === -2) {
				$dead++;
				$class = 'dead';
			}
			if ($output_prefix !== false) {
				$line = htmlspecialchars($line, ENT_QUOTES, 'UTF-8');
				$line = str_replace(' ', '&nbsp;', $line);
				$line = str_replace("\t", '&nbsp;&nbsp;&nbsp;&nbsp;', $line);
				fwrite($out, '<li class="'.$class.'">'.$line.'</li>');
			}
			$i++;
		}
		// XXX: It seems like (number of lines in file) + 1 is always set, which throws of our percentages.
		if (isset($lines[$i]))
			unset($lines[$i]);
		$nlines = count($lines);
		$row = sprintf ('%s: %d executable lines, %.2f%% coverage, %.2f%% unused, %.2f%% dead', $file, $nlines, $visited * 100/$nlines, $unused * 100/$nlines, $dead * 100/$nlines);
		// only print file data where it can be improved
		if (!$unused)
			$completely_tested[] = $file;
		if ($output_prefix !== false) {
			fwrite($out, '</ol>');
			fwrite($out, '<p>'.$row.'</p>');
			fwrite($out, '</body></html>');
			fclose($out);
			fclose($in);
		}
		$total_lines += $nlines;
		$total_covered += $visited;
		$all_files[$file] = array('nlines' => $nlines, 'visited' => $visited, 'unused' => $unused, 'dead' => $dead);
	}
	$coverage = 100 * $total_covered / $total_lines;
	
	printf('%d tested lines out of %d = %.2f%% coverage | lines=%2$d;;;; covered=%1$d;;;; percentage=%3$f%%;%4$d;%5$d;;', $total_covered, $total_lines, $coverage, $warning, $critical);
	print "\n";
	foreach ($all_files as $file => $data)
		printf ("%s: %d executable lines, %.2f%% coverage, %.2f%% unused, %.2f%% dead\n", $file, $data['nlines'], $data['visited'] * 100/$data['nlines'], $data['unused'] * 100/$data['nlines'], $data['dead'] * 100/$data['nlines']);
	if (!empty($completely_tested))
		print "Completely tested: ".implode(', ', $completely_tested)."\n";
		
	$total = sprintf('Total: %d out of %d = %.2f%%', $total_covered, $total_lines, $coverage);
	if ($output_prefix !== false) {
		$h = fopen($output_prefix.'/index.html', 'w');
		fwrite($h, '<html><head><script src="jquery.js"></script><script src="jquery.dataTables.min.js"></script><script>$(document).ready(function() {$("table").dataTable({bPaginate:false});});</script><style>th {cursor:pointer;}</style></head><body>');
		fwrite($h, '<p>'.$total.'</p>');
		fwrite($h, '<div><table><thead><tr><th>File</th><th>Lines</th><th>Coverage</th><th>Unused</th><th>Dead</th></tr></thead><tbody>');
		foreach ($all_files as $file => $data) {
			$row = sprintf('<tr><td><a href="%s.html">%s</td><td>%d</td><td>%.2f%%</td><td>%.2f%%</td><td>%.2f%%</td></tr>', $file, $file, $data['nlines'], $data['visited'] * 100/$data['nlines'], $data['unused'] * 100/$data['nlines'], $data['dead'] * 100/$data['nlines']);
			fwrite($h, $row);
		}
		fwrite($h, '</tbody></table></div></body></html>');
		fclose($h);
		copy(dirname(__FILE__).'/jquery.dataTables.min.js', $output_prefix.'/jquery.dataTables.min.js');
		copy(dirname(__FILE__).'/jquery.js', $output_prefix.'/jquery.js');
	}

	if (is_numeric($critical) && $coverage < $critical)
		return 2;
	else if (is_numeric($warning) && $coverage < $warning)
		return 1;
	else if (!is_numeric($critical) && !is_numeric($warning))
		return 3;
	else
		return 0;
}
