<?php
require_once('phptap.php');

/**
 * A fooUnit like base class for PHPTap
 *
 * To use, inherit from it, and write your tests in methods named test*.
 * TapUnit will create one sub-tap for each class, and one for each method, so
 * make sure to give them good names.
 *
 * You can make per-test initialization/deinitialization by overriding setUp
 * and tearDown, and you can make per-test class
 * initialization/deinitialization in the static methods
 * setUpClass/tearDownClass.
 *
 * You can hide output in the tests by setting the static variable
 * $useOutputBuffering to true.
 *
 * You can not use the javaesque, novel-length assert functions that are
 * usually present in fooUnit classes: you'll have to make do with the
 * easy-to-type ones avalible in phptap.
 */
class TapUnit {
	function setUp() {}
	function tearDown() {}
	static function setUpClass() {}
	static function tearDownClass() {}
	var $useOutputBuffering = false;
	private $tap = null;
	private $main_tap = null;

	function __construct($tap)
	{
		$this->main_tap = $tap;
	}

	function runTests()
	{
		$vars = get_class_methods($this);
		foreach ($vars as $func) {
			if (substr($func, 0, 4) === 'test') {
				$this->tap = $this->main_tap->sub_init($func);
				$this->tap->files_to_skip = array(__FILE__);
				if ($this->useOutputBuffering)
					ob_start();
				try {
					$this->setUp();
					$this->$func();
					$this->tearDown();
				} catch (Exception $ex) {
					$this->tap->fail('Unhandled exception: '.$ex->getMessage());
				}
				if ($this->useOutputBuffering)
					ob_end_clean();
				$this->tap->done();
			}
		}
	}

	// These could all be replaced by a __call method, but that causes the
	// phptap output to look terrible.

	function ok($result, $msg, $flags=0)
	{
		if ($this->useOutputBuffering)
			ob_end_clean();
		$this->tap->ok($result, $msg, $flags);
		if ($this->useOutputBuffering)
			ob_start();
	}

	function ok_eq($a, $b, $msg, $flags=0)
	{
		if ($this->useOutputBuffering)
			ob_end_clean();
		$this->tap->ok_eq($a, $b, $msg, $flags);
		if ($this->useOutputBuffering)
			ob_start();
	}

	function ok_id($a, $b, $msg, $flags=0)
	{
		if ($this->useOutputBuffering)
			ob_end_clean();
		$this->tap->ok_id($a, $b, $msg, $flags);
		if ($this->useOutputBuffering)
			ob_start();
	}

	function ok_empty($ary, $msg, $flags=0)
	{
		if ($this->useOutputBuffering)
			ob_end_clean();
		$this->tap->ok_empty($ary, $msg, $flags);
		if ($this->useOutputBuffering)
			ob_start();
	}

	function pass($msg, $flags=0)
	{
		if ($this->useOutputBuffering)
			ob_end_clean();
		$this->tap->pass($msg, $flags);
		if ($this->useOutputBuffering)
			ob_start();
	}

	function fail($msg, $flags=0)
	{
		if ($this->useOutputBuffering)
			ob_end_clean();
		$this->tap->fail($msg, $flags);
		if ($this->useOutputBuffering)
			ob_start();
	}

	function diag($msg)
	{
		if ($this->useOutputBuffering)
			ob_end_clean();
		$this->tap->fail($msg);
		if ($this->useOutputBuffering)
			ob_start();
	}
}

/**
 * Runs a tapunit derived class
 * (this would have been a static method if PHP 5.2 would have had
 * get_called_class
 */
function runUnit($class_name, $tap)
{
	$sub = $tap->sub_init($class_name);
	$obj = new $class_name($sub);
	$obj->setUpClass();
	$obj->runTests();
	$obj->tearDownClass();
	$sub->done();
}

?>
